(* {{{ COPYING *)
(*
 * +---------------------------------------------------------------+
 * | Copyright (C) 2006 Florent Monnier                            |
 * +---------------------------------------------------------------+
 * | This is a small quick and dirty starting point for X demos.   |
 * +---------------------------------------------------------------+
 * |                                                               |
 * | This program is free software; you can redistribute it and/or |
 * | modify it under the terms of the GNU General Public License   |
 * | as published by the Free Software Foundation; either version  |
 * | 2 of the License, or (at your option) any later version.      |
 * |                                                               |
 * | This program is distributed in the hope that it will be       |
 * | useful, but WITHOUT ANY WARRANTY; without even the implied    |
 * | warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       |
 * | PURPOSE. See the GNU General Public License for more details. |
 * |                                                               |
 * | http://www.fsf.org/licensing/licenses/gpl.html                |
 * |                                                               |
 * | You should have received a copy of the GNU General Public     |
 * | License along with this program; if not,                      |
 * | write to the Free Software Foundation, Inc.,                  |
 * | 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA  |
 * |                                                               |
 * +---------------------------------------------------------------+
 * | Author: Florent Monnier <fmonnier@linux-nantes.fr.eu.org>     |
 * +---------------------------------------------------------------+
 *
 * }}} *)

(* {{{ init vars *)

open StdLabels ;;

let width, height  = 1024, 768 ;;
let total_seconds  = 90 ;;
let frames_per_sec = 60 ;;

(* }}} *)

(* {{{ utilities *)

type init_item = Width | Height | Frames_per_sec | Total_seconds ;;

let init item =
  match item with
  | Width  -> width
  | Height -> height
  | Frames_per_sec -> frames_per_sec
  | Total_seconds -> total_seconds
;;

external set_item: init_item -> int -> unit = "ml_set_item";;


external xSet_foreground: index:int -> red:int -> green:int -> blue:int -> unit = "ml_xsetforeground";;
let xSet_main_foreground ~red ~green ~blue = xSet_foreground ~index:0 ~red ~green ~blue;;


external xChangeGC_line_width: index:int -> line_width:int -> unit = "ml_xchangegc_foreground_i_line_width"
external xChangeGC_background_line_width:  line_width:int -> unit  = "ml_xchangegc_background_line_width"

(* how to composite with the destination pixel *)
type func =
  | GXclear         (*  0x0   0                       *)
  | GXand           (*  0x1   src AND dst             *)
  | GXandReverse    (*  0x2   src AND NOT dst         *)
  | GXcopy          (*  0x3   src                     *)
  | GXandInverted   (*  0x4   (NOT src) AND dst       *)
  | GXnoop          (*  0x5   dst                     *)
  | GXxor           (*  0x6   src XOR dst             *)
  | GXor            (*  0x7   src OR dst              *)
  | GXnor           (*  0x8   (NOT src) AND (NOT dst) *)
  | GXequiv         (*  0x9   (NOT src) XOR dst       *)
  | GXinvert        (*  0xa   NOT dst                 *)
  | GXorReverse     (*  0xb   src OR (NOT dst)        *)
  | GXcopyInverted  (*  0xc   NOT src                 *)
  | GXorInverted    (*  0xd   (NOT src) OR dst        *)
  | GXnand          (*  0xe   (NOT src) OR (NOT dst)  *)
  | GXset           (*  0xf   1                       *)
;;

external xChangeGC_function: index:int -> func:func -> unit = "ml_xchangegc_foreground_i_function";;
external xChangeGC_background_function: func:func -> unit = "ml_xchangegc_background_function";;

let string_of_func ~func = match func with
  | GXclear        -> "GXclear"
  | GXand          -> "GXand"
  | GXandReverse   -> "GXandReverse"
  | GXcopy         -> "GXcopy"
  | GXandInverted  -> "GXandInverted"
  | GXnoop         -> "GXnoop"
  | GXxor          -> "GXxor"
  | GXor           -> "GXor"
  | GXnor          -> "GXnor"
  | GXequiv        -> "GXequiv"
  | GXinvert       -> "GXinvert"
  | GXorReverse    -> "GXorReverse"
  | GXcopyInverted -> "GXcopyInverted"
  | GXorInverted   -> "GXorInverted"
  | GXnand         -> "GXnand"
  | GXset          -> "GXset"
;;


external xClearArea: x:int -> y:int -> width:int -> height:int -> unit = "ml_xcleararea";;
external xClearWindow: unit -> unit = "ml_xclearwindow";;
external xClear_buffer: unit -> unit = "ml_xclear_buffer";;

external xCloseDisplay: unit -> unit = "ml_xclosedisplay";;

external xCopyArea: unit -> unit = "ml_xcopyarea"

external usleep: int -> unit = "ml_usleep"

type graphic_context = Background | Main_Foreground | Foreground of int

external xDrawLine_foreground: index:int -> x1:int -> y1:int -> x2:int -> y2:int -> unit = "ml_xdrawline_foreground_i";;
external xDrawLine_main_foreground: x1:int -> y1:int -> x2:int -> y2:int -> unit = "ml_xdrawline_main_foreground";;
external xDrawLine_background:      x1:int -> y1:int -> x2:int -> y2:int -> unit = "ml_xdrawline_background";;
external xDrawLine_buffer:          x1:int -> y1:int -> x2:int -> y2:int -> unit = "ml_xdrawline_buffer";;


external xDrawRectangle_background:      x:int -> y:int -> width:int -> height:int -> unit = "ml_xdrawrectangle_background";;
external xDrawRectangle_main_foreground: x:int -> y:int -> width:int -> height:int -> unit = "ml_xdrawrectangle_main_foreground";;
external xDrawRectangle_foreground: index:int -> x:int -> y:int -> width:int -> height:int -> unit = "ml_xdrawrectangle_foreground_i";;

let xDrawLine ?(color=Main_Foreground) ~x1 ~y1 ~x2 ~y2 () =
  match color with
  | Background       -> xDrawLine_background      ~x1 ~y1 ~x2 ~y2
  | Main_Foreground  -> xDrawLine_main_foreground ~x1 ~y1 ~x2 ~y2
  | Foreground index -> xDrawLine_foreground      ~x1 ~y1 ~x2 ~y2 ~index
;;

let xDrawRectangle ~color =
  assert (width  > 0);
  assert (height > 0);
  match color with
  | Background       -> xDrawRectangle_background
  | Main_Foreground  -> xDrawRectangle_main_foreground
  | Foreground index -> xDrawRectangle_foreground ~index
;;

let xDrawRectangle' ?(color=Main_Foreground) ~x1 ~y1  ~x2 ~y2 () =
  let (x, width)  = if x1 < x2 then (x1, x2 - x1) else (x2, x1 - x2)
  and (y, height) = if y1 < y2 then (y1, y2 - y1) else (y2, y1 - y2)
  in
  match color with
  | Background       -> xDrawRectangle_background      ~x ~y  ~width ~height;
  | Main_Foreground  -> xDrawRectangle_main_foreground ~x ~y  ~width ~height;
  | Foreground index -> xDrawRectangle_foreground      ~x ~y  ~width ~height ~index;
;;


external xDrawPoint_main_foreground: x:int -> y:int -> unit = "ml_xdrawpoint_main_foreground";;
external xDrawPoint_foreground: index:int -> x:int -> y:int -> unit = "ml_xdrawpoint_foreground_i";;


type points_list = (int * int) list ;;

external _xDrawPoints_main_foreground: points_list:points_list -> npoints:int -> unit = "ml_xdrawpoints_main_foreground";;
let xDrawPoints_main_foreground ~points =
  let npoints = List.length points in
  _xDrawPoints_main_foreground ~points_list:points ~npoints
;;

external _xDrawPoints_foreground: index:int -> points_list:points_list -> npoints:int -> unit = "ml_xdrawpoints_foreground_i";;
let xDrawPoints_foreground ~index ~points =
  let npoints = List.length points in
  _xDrawPoints_foreground ~index ~points_list:points ~npoints
;;

type shape =
  | Convex
  | NonConvex
  | Complex
;;

external _xFillPolygon_foreground: index:int -> points_list:points_list -> npoints:int -> shape:shape -> unit
    = "ml_xfillpolygon_foreground_i"
let xFillPolygon_foreground ~index ~points ?(shape=Complex) () =
  let npoints = List.length points in
  _xFillPolygon_foreground ~index ~points_list:points ~npoints ~shape
;;
let xFillPolygon ?(index=0) ~points ?(shape=Complex) () =
  xFillPolygon_foreground ~index ~points ~shape ()
;;


external xDrawArc_background: x:int -> y:int -> width:int -> height:int -> angle1:int -> angle2:int -> unit
    = "ml_xdrawarc_background_bytecode"
      "ml_xdrawarc_background_native" ;;

external xDrawArc_foreground: index:int -> x:int -> y:int -> width:int -> height:int -> angle1:int -> angle2:int -> unit
    = "ml_xdrawarc_foreground_i_bytecode"
      "ml_xdrawarc_foreground_i_native" ;;

let xDrawArc_main_foreground = xDrawArc_foreground ~index:0 ;;

let xDrawArc ~color =
  match color with
  | Background       -> xDrawArc_background
  | Main_Foreground  -> xDrawArc_main_foreground
  | Foreground index -> xDrawArc_foreground ~index
;;

type coords_2d = {x:int; y:int} ;;

let xDrawArc' ?(color=Main_Foreground) ~center ~radius ~angle1 ~angle2 () =
  let x = center.x - radius.x
  and y = center.y - radius.y
  and width  = 2 * radius.x
  and height = 2 * radius.y
  in
  xDrawArc ~color ~x ~y ~width ~height ~angle1 ~angle2
;;

let xDrawElipse ?(color=Main_Foreground) ~center ~radius () =
  xDrawArc' ~color ~center ~radius ~angle1:0 ~angle2:23040 ()  (* 360 * 64 = 23040 *)
;;

let xDrawCircle ?(color=Main_Foreground) ~center ~radius () =
  xDrawArc' ~color ~center ~radius:{x=radius; y=radius} ~angle1:0 ~angle2:23040 ()  (* 360 * 64 = 23040 *)
;;


type t_arc =
  { x_arc  : int;
    y_arc  : int;
    width  : int;
    height : int;
    angle1 : int;
    angle2 : int; }
;;
external _xDrawArcs_foreground: index:int -> arcs_list:t_arc list -> narcs:int -> unit = "ml_xdrawarcs_foreground_i";;
let xDrawArcs_foreground ~index ~arcs_list =
  let narcs = List.length arcs_list in
  _xDrawArcs_foreground ~index ~arcs_list ~narcs
;;


external xFlush: unit -> unit = "ml_xflush"


(* }}} *)

#define FRAME_BUFFER 0

(* {{{ iteration loop *)

let screen_dist = 120.0 (* distance from viewpoint to screen *)

let c1 = 60.0
let c2 = 40.0

let f1 = 8.0
let f2 = 8.0

(* {{{ coef *)

let coef =
  let factor = (atan 1.0) /. 45.0 in
  (fun ~rho ~theta ~phi ->
    (* angles in radians *)
    let th = factor *. theta
    and ph = factor *. phi
    in
    let costh = cos th and sinth = sin th
    and cosph = cos ph and sinph = sin ph
    in
    (* elements of matrix *)
    let v11 = (-.sinth)
    and v12 = (-.cosph) *. costh
    and v13 = (-.sinph) *. costh
    and v21 = costh
    and v22 = (-.cosph) *. sinth
    and v23 = (-.sinph) *. sinth
    and v32 = sinph
    and v33 = (-.cosph)
    and v43 = rho
    in
    ( v11, v12, v13,
      v21, v22, v23,
           v32, v33,
                v43 )
  )
;;
(* }}} *)
(* {{{ perspective *)

let perspective x y z ~mat:(v11, v12, v13,
                            v21, v22, v23,
                                 v32, v33,
                                      v43 ) =
  (* eye coordinates *)
  let xe = v11 *. x +. v21 *. y
  and ye = v12 *. x +. v22 *. y +. v32 *. z
  and ze = v13 *. x +. v23 *. y +. v33 *. z +. v43
  in
  (* screen coordinantes *)
  let x = screen_dist *. xe /. ze
  and y = screen_dist *. ye /. ze in
  let x = f1 *. (x +. c1)
  and y = f2 *. (y +. c2) in
  (truncate x, truncate y)
;;
(* }}} *)
(* {{{ draw segmets *)

let draw_segment ~mat ~pt1:(x1, y1, z1) ~pt2:(x2, y2, z2) =
  let x1, y1 = perspective x1 y1 z1 ~mat
  and x2, y2 = perspective x2 y2 z2 ~mat
  in
#if FRAME_BUFFER
  xDrawLine_buffer ~x1 ~y1 ~x2 ~y2;
#else
  xDrawLine ~x1 ~y1 ~x2 ~y2 ();
#endif
;;

let draw_segments ~mat ~segments =
  let rec aux pt1 tl_pt =
    match tl_pt with
    | [] -> ()
    | pt2::tail ->
        draw_segment ~mat ~pt1 ~pt2;
        aux pt2 tail
  in
  match segments with
  | first_point::tail -> aux first_point tail
  | [] -> ()
;;
(* }}} *)

at_exit (fun () -> print_newline())

(* {{{ dse reader *)
(* {{{ utilities *)
let get_points i =
  let rec aux acc =
    try let ln = input_line i in
      match ln with
      | "" -> List.rev acc
      | _  -> aux (ln::acc)
    with End_of_file -> failwith "Error: no separator"
  in
  aux []
;;

let get_meshes i =
  let rec aux acc =
    try let ln = input_line i in
      aux (ln::acc)
    with End_of_file -> close_in i; List.rev acc
  in
  aux []
;;

let point_sep p =
  let coords = Str.split (Str.regexp_string " ") p in
  List.map (fun coord -> float_of_string coord) coords
;;

let meshes_sep points m =
  let indexed = Str.split (Str.regexp_string " ") m in
  let deref i =
    let i = int_of_string i in
    points.(i)
  in
  let coords = List.map deref indexed in
  coords
;;
(* }}} *)

let pos (x, y, z) =
  (x *. 1000.0, y *. (-.1000.0), (z *. (-.1000.0)) +. 1000.0)
;;

let load_file ~filename =
  let i = open_in filename in
  let points = get_points i in
  let points = List.map point_sep points in
  let points = Array.of_list points in
  let meshes = get_meshes i in
  let meshes = List.map (meshes_sep points) meshes in
  (* {{{ meshes *)
  let meshes =
    List.map (fun mesh ->
      List.map (fun point ->
        let p_ar = Array.of_list point in
        pos (p_ar.(0), p_ar.(1), p_ar.(2))
      ) mesh;
    ) meshes
  in
  (* }}} *)
  meshes
;;

let face = load_file ~filename:"face.dse" ;;
let ears = load_file ~filename:"ears.dse" ;;
(*
let skull = load_file ~filename:"skull.dse" ;;
*)

(* }}} *)

let rec main_loop i =
  let rec aux i =
    (*
    Printf.printf " %d" i;
    *)
#if FRAME_BUFFER
    xClear_buffer();
#else
    xClearWindow();
#endif
    xSet_main_foreground ~red:28000 ~green:52000 ~blue:65500;
    let rho   = 3200.0  (* viewing distance OE *)
    and theta =  30.0 +. (0.40 *. float i)  (* measured horizontally from the x-axis *)
    and phi   =  70.0 +. (0.04 *. float i)  (* measured verticaly from the z-axis *)
    in
    (* {{{ *(
    for j = 0 to 5 do
      let k = (float j) *. 25.0 in
      let h = 400.0 +. k in
      let mat = coef ~rho ~theta ~phi in
      let put = draw_segments ~mat in
      let segments = [
        [ (h, -.h, -.h); (h, h, -.h); (-.h, h, -.h);
          (-.h, h, h); (-.h, -.h, h); (h, -.h, h); (h, -.h, -.h) ];
        [ (h, h, -.h); (h, h, h); (-.h, h, h);
          (-.h, -.h, h); (h, -.h, h); (h, -.h, -.h) ];
        [ (h, h, -.h); (h, h, h); (-.h, h, h) ];
        [ (h, h, h); (h, -.h, h) ];
        [ (h, -.h, -.h); (-.h, -.h, -.h); (-.h, h, -.h) ];
        [ (-.h, -.h, -.h); (-.h, -.h, h) ];
      ] in
      (* {{{ translate *(
      let translate segments =
        List.map (fun (x,y,z) ->
          (x +. 0.0),
          (y +. 0.0),
          (z)
        ) segments
      in
      let segments = List.map translate segments in
      )* }}} *)
      List.iter (fun segments -> put ~segments) segments;
    done;
    )* }}} *)

    let mat = coef ~rho ~theta ~phi in
    let put = draw_segments ~mat in
    List.iter (fun segments -> put ~segments) face;
    xSet_main_foreground ~red:4000 ~green:24000 ~blue:40000;
    List.iter (fun segments -> put ~segments) ears;
    (*
    List.iter (fun segments -> put ~segments) skull;
    *)

#if FRAME_BUFFER
    xCopyArea();
#endif
    xFlush();

    if i < 4600 then aux (succ i)
    (*
    else Printf.printf "\n%!";
    *)
  in
  aux i
;;

(* }}} *)

(* {{{ main *)

let _ =
  Callback.register "ocaml init" init;
  Callback.register "ocaml main loop" main_loop;
;;

(* }}} *)


(* vim: sw=2 sts=2 ts=2 et fdm=marker
 *)
